package com.mxnavi.tspv2.dns;

import com.mxnavi.tspv2.dns.model.DomainInfo;
import com.mxnavi.tspv2.dns.model.IpResult;
import com.mxnavi.tspv2.core.Config;
import com.mxnavi.tspv2.utils.Utils;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by wangq on 2016/6/7.
 * 域名解析缓存服务类
 */
public class DnsService {
    static DnsService dnsService;
    public synchronized static DnsService getInstance(){
        if(dnsService == null){
            dnsService =  new DnsService();
//            Security.setProperty("networkaddress.cache.ttl",
//                    String.valueOf(0));
//
//            Security.setProperty("networkaddress.cache.negative.ttl",
//                    String.valueOf(0));
        }
        return dnsService;
    }
    ConcurrentHashMap<String,DomainInfo> domainInfoHashMap = new ConcurrentHashMap<String,DomainInfo>();
    // 最小过期时间，为每次循环更新间歇时间
    long minTimeOuts = 0;
    // DNS解析域名过期时间为5分钟
    int dnsTTL = 300;

    final int STATE_NOTEXIST = 0;
    final int STATE_PARSING = 1;
    final int STATE_PARSED = 2;
    final int STATE_EXPIRE = 3;
    final int PARSETYPE_DPLUS = 1;
    final int PARSETYPE_NDS =2;
    final int PARSETYPE_ALL =3;
    volatile boolean isParse = true;
    Thread parseThread = null;
    Object  object = new Object();

    /**
     * 通过域名同步获取ip
     * @param domain 域名
     * @return
     */
    public IpResult getIpByDomain(String domain){
        DomainInfo domainInfo =  domainInfoHashMap.get(domain);
        IpResult ipResult = new IpResult();
        if(domainInfo==null){
            ipResult.setState(STATE_NOTEXIST);
            return ipResult;
        }else{
            if(Utils.strIsEmpty(domainInfo.httpPodDnsIp)) {
                if (Utils.strIsEmpty(domainInfo.dnsIp)) {
                    ipResult.setState(STATE_PARSING);
                }else{
                    if( isExpire(domainInfo)){
                        // 已超时
                        ipResult.setState(STATE_EXPIRE);
                    }else{
                        ipResult.setState(STATE_PARSED);
                    }
                    ipResult.setIp(domainInfo.dnsIp);

                }
            }else{
                ipResult.setIp(domainInfo.httpPodDnsIp);

                if(isExpire(domainInfo)){
                    // 已超时
                    ipResult.setState(STATE_EXPIRE);
                }else{
                    ipResult.setState(STATE_PARSED);
                }
            }
        }
        return ipResult;
    }

    /**
     * 添加要解析的域名，异步返回ip
     * @param domain 域名
     * @param parseType 解析方式：1-只用D+解析；2-只用DNS解析；3-D+和DNS都可以用，优先选择D+
     * @param onGetIpListener 解析成功后回调接口，不能为null
     */
    public synchronized void addDomainName(String domain,int parseType,OnGetIpListener onGetIpListener){
        if(onGetIpListener==null){
            return ;
        }
        DomainInfo domainInfo =  domainInfoHashMap.get(domain);
        if(domainInfo!=null){
            switch (parseType){
                case PARSETYPE_ALL:// D+和DNS都解析
                    if(!Utils.strIsEmpty(domainInfo.httpPodDnsIp)){
                        onGetIpListener.onGetIp(domainInfo.httpPodDnsIp);
                    }else if(!Utils.strIsEmpty(domainInfo.dnsIp)){
                        onGetIpListener.onGetIp(domainInfo.dnsIp);
                    }else{
                        if(!domainInfo.OnGetIpListeners.contains(onGetIpListener)){
                            domainInfo.OnGetIpListeners.add(onGetIpListener);
                        }
                    }
                    break;
                case PARSETYPE_DPLUS:// 只需要D+解析
                    if(!Utils.strIsEmpty(domainInfo.httpPodDnsIp)){
                        onGetIpListener.onGetIp(domainInfo.httpPodDnsIp);
                    }else{
                        if(!domainInfo.OnGetIpListeners.contains(onGetIpListener)){
                            domainInfo.OnGetIpListeners.add(onGetIpListener);
                        }
                    }
                    break;
                case PARSETYPE_NDS:// 只需要DNS解析
                    if(!Utils.strIsEmpty(domainInfo.dnsIp)){
                        onGetIpListener.onGetIp(domainInfo.dnsIp);
                    }else{
                        if(!domainInfo.OnGetIpListeners.contains(onGetIpListener)){
                            domainInfo.OnGetIpListeners.add(onGetIpListener);
                        }
                    }
                    break;
            }

        }else{

            domainInfo = new DomainInfo();
            domainInfo.domain = domain;
            domainInfo.OnGetIpListeners.add(onGetIpListener);
            domainInfo.state = STATE_PARSING;
            domainInfo.parseWay = parseType;
            domainInfoHashMap.put(domain,domainInfo);
            if(parseThread == null){
                parseThread = new Thread(new DNSParse(),"TSP-SDK-DNSParse-thread");
                parseThread.start();
            }else{
                synchronized (object) {
//                    Utils.loge("唤醒");
                    object.notify();
                }
            }
        }
    }
    /**
     * 结束域名解析wait，马上解析
     */
    public void notifyParseDomain(){
        synchronized (object) {
//          Utils.loge("唤醒");
          object.notify();
      }
    }
    public void  stopDnsService(){
        isParse = false;
        dnsService = null;
    }

    class DNSParse implements Runnable{
        private String pareseDNS( String domainName){
            String ip = null;
            try {
                InetAddress myServer = InetAddress.getByName(domainName);
                ip = myServer.getHostAddress();
                Utils.logd(domainName + "-DNS解析结果：" + ip);
            } catch (UnknownHostException e) {
                e.printStackTrace();
            }catch(Exception e){
                e.printStackTrace();
            }
            return ip;
        }
        private String parseHttpPodDns( String domainName){
            HttpURLConnection connection = null;
            InputStream inputStream;
            String strurl =  Config.DPlUS_URL+domainName;
            URL url = null;
            try {
//                Utils.loge("strurl"+strurl);
                url = new URL(strurl);
                connection = (HttpURLConnection) url.openConnection();
                connection.setConnectTimeout(15000);
                connection.setRequestMethod("GET");
                connection.setReadTimeout(15000);
                connection.setUseCaches(false);
                connection.addRequestProperty("User-Agent", "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.94 Safari/537.36");

                // 获得响应状态
                int responseCode = connection.getResponseCode();
                inputStream = connection.getInputStream();
                Utils.logd(domainName+"-D+解析responseCode：" + responseCode);
                if (HttpURLConnection.HTTP_OK == responseCode) {// 连接成功
                    StringBuffer sb = new StringBuffer();
                    String readLine;
                    BufferedReader responseReader;
                    responseReader = new BufferedReader(new InputStreamReader(inputStream));
                    while ((readLine = responseReader.readLine()) != null) {
                        sb.append(readLine);
                    }
                    responseReader.close();
                    String result = sb.toString();
                    Utils.logd(domainName + "-D+解析结果：" + result);
                    String ipStr = result.split(",")[0];
                    String[] ips = ipStr.split(";");
                    if(Utils.isIP(ips[0])){
                        return result;
                    }else{
                        return null;
                    }


                }else {
                    Utils.logd(domainName + "-D+解析结果2：null");
                    return null;
                }
            }catch (MalformedURLException e) {
                e.printStackTrace();
                return null;
            }catch (IOException e) {
                e.printStackTrace();
                return null;
            } catch (Exception e) {
                e.printStackTrace();
                return null;
            }
        }
        public void onParseDomainDone(String ip,DomainInfo domainInfo,int type){
            String ipStr = ip.split(",")[0];
            String[] ips = ipStr.split(";");
            String returnIp = "";
            Utils.logd("解析出的ip=" + ip);
            if(type == 0){// d+解析
                if(!Utils.strIsEmpty(domainInfo.httpPodDnsIp)){// 原d+Ip不为空，则比较新解析出的IP跟原IP是否相同，如果不相同，则替换
                    if(!ip.contains(domainInfo.httpPodDnsIp)){
                        domainInfo.httpPodDnsIp = ips[0];
                    }
                }else{// 原D+Ip为空，则直接赋给ip列表中的第一个值
                    domainInfo.httpPodDnsIp = ips[0];
                }
                try {
                    domainInfo.ttl = Long.parseLong(ip.split(",")[1]);
                }catch (Exception e){
                    e.printStackTrace();
                    domainInfo.ttl = dnsTTL;
                }

                returnIp =  domainInfo.httpPodDnsIp;
            }else{
                if(!Utils.strIsEmpty(domainInfo.dnsIp)){// 原dnsIp不为空，则比较新解析出的IP跟原IP是否相同，如果不相同，则替换
                    if(!ip.contains(domainInfo.dnsIp)){
                        domainInfo.dnsIp = ips[0];
                    }else{
                    }
                }else{
                    domainInfo.dnsIp = ips[0];
                }
                returnIp =  domainInfo.dnsIp;
                domainInfo.ttl = dnsTTL;
            }

            domainInfo.parseTime = System.currentTimeMillis();
            domainInfo.state = STATE_PARSED;
            for(OnGetIpListener onGetIpListener :domainInfo.OnGetIpListeners){
//                onGetIpListener.OnGetIp(onGetIpListener.toString()+"-"+domainInfo.domain);
                onGetIpListener.onGetIp(returnIp);
            }
            domainInfo.OnGetIpListeners.clear();


        }
        public void run(){
            try{

                while(isParse){
                    Iterator<Map.Entry<String, DomainInfo>> entries = domainInfoHashMap.entrySet().iterator();
                    int length = domainInfoHashMap.size();
//                    Utils.loge("domainInfoHashMap.size="+length);
                    int parseFailNum = 0;
                    int i = 0;
                    while (entries.hasNext()&&isParse) {
                        Map.Entry<String, DomainInfo> entry = entries.next();
                        String domain = entry.getKey();
                        DomainInfo domainInfo = entry.getValue();
//                        Utils.loge("循环key="+domain+"; DomainInfo="+domainInfo.toString());
                        switch (domainInfo.parseWay) {
                            case PARSETYPE_ALL:
                                if(((!Utils.strIsEmpty(domainInfo.dnsIp)||!Utils.strIsEmpty(domainInfo.httpPodDnsIp))
                                        &&isExpire(domainInfo))
                                        ||(Utils.strIsEmpty(domainInfo.dnsIp)&&Utils.strIsEmpty(domainInfo.httpPodDnsIp))){

                                    // 先D+,如果失败再DNS
                                    String ip= parseHttpPodDns(domain);
                                    if(!Utils.strIsEmpty(ip)){// D+
                                        onParseDomainDone(ip,domainInfo,0);

                                    }else{//D+解析失败,进行DNS解析
                                        ip = pareseDNS(domain);
                                        if(!Utils.strIsEmpty(ip)){
                                            onParseDomainDone(ip,domainInfo,1);
                                        }else{// 解析失败
                                            parseFailNum ++;
                                        }
                                    }
                                }
                                break;
                            case PARSETYPE_DPLUS:// 只使用D+
                                if(Utils.strIsEmpty(domainInfo.httpPodDnsIp)|| isExpire(domainInfo)){
                                    // D+解析
                                    // 重新解析，先D+,如果失败再DNS
                                    String ip = parseHttpPodDns(domain);
                                    if(!Utils.strIsEmpty(ip)){// D+
                                        onParseDomainDone(ip,domainInfo,0);

                                    }else {//D+解析失败
                                        parseFailNum ++;
                                    }

                                }
                                break;
                            case PARSETYPE_NDS:// 只进行DNS解析
                                if(Utils.strIsEmpty(domainInfo.dnsIp)|| isExpire(domainInfo)){
                                    // DNS解析
                                    String ip = pareseDNS(domain);
                                    if(!Utils.strIsEmpty(ip)){
                                        onParseDomainDone(ip,domainInfo,1);
                                    }else{// 解析失败
                                        parseFailNum ++;
                                    }

                                }
                                break;
                        }

                        if(i==0){
                            minTimeOuts = domainInfo.ttl;

                        }else{
                            if(minTimeOuts==0||minTimeOuts> (domainInfo.ttl)){
                                minTimeOuts =  domainInfo.ttl;
                            }
                        }
                        i++;
                    }
//                    Utils.loge("domainInfoHashMap.size新="+domainInfoHashMap.size());
                    if(parseFailNum == 0&&length ==domainInfoHashMap.size()){
                        synchronized (object) {
                            try {
                                Utils.logd("循环时间 minTimeOuts=" + minTimeOuts);
                                if(minTimeOuts<=0){
                                    minTimeOuts = dnsTTL;
                                }
                                object.wait(minTimeOuts*1000);
                            } catch (InterruptedException e) {
                                e.printStackTrace();
                            }
                        }

                    }else if(parseFailNum>0){
                        synchronized (object) {
                            try {
                                object.wait(60*1000);
                            } catch (InterruptedException e) {
                                e.printStackTrace();
                            }
                        }
                    }
                }

            }catch (Exception e){
                e.printStackTrace();
            }
        }
    }
//    public void test(){
//        Thread t1 = new th("www.baidu.com");
//        Thread t2 = new th("mx-t10-longlived-v1.mxnavi.com");
//        Thread t3 = new th("my.oschina.net");
//        Thread t4 = new th("www.qq.com");
//        Thread t5 = new th("www.sina.com");
//        t1.start();
//        t2.start();
//        t3.start();
//        t4.start();
//        t5.start();
//    }
//    class th extends Thread {
//        private final String mkey ;
//
//        public th(String key) {
//            mkey = key;
//        }
//
//        public void run() {
//            if("www.qq.com".equals(mkey)){
//                try {
//                    Thread.currentThread().sleep(8000);
//                } catch (InterruptedException e) {
//                    e.printStackTrace();
//                }
//            }
//            if("www.sina.com".equals(mkey)){
//                try {
//                    Thread.currentThread().sleep(9000);
//                } catch (InterruptedException e) {
//                    e.printStackTrace();
//                }
//            }
//            for (int i = 0; i < 10; i++) {
//                final int k = i;
//                OnGetIpListener OnGetIpListener1 = new OnGetIpListener() {
//                    public void OnGetIp(String ip) {
//                        Utils.loge("回调-"+k+"IP==" + ip+"-mkey="+mkey);
//                    }
//                };
//                Utils.loge(mkey + "添加" + OnGetIpListener1.toString());
//                addDomainName(mkey, 3, OnGetIpListener1);
//                Utils.loge("getIpByDomain"+getIpByDomain(mkey).toString());
//                try {
//                    Thread.currentThread().sleep(500);
//                } catch (InterruptedException e) {
//                    e.printStackTrace();
//                }
//
//            }
//        }
//    }

    private boolean isExpire(DomainInfo domainInfo) {
        if((domainInfo.parseTime+(domainInfo.ttl)*1000)<=System.currentTimeMillis()){
            return true;
        }
       return false;
    }





}
